#define CCC1(x) #x
#define CCC2(x) CCC1(x)
#define CCC { printk(__FUNCTION__ " (" __FILE__ ":" CCC2(__LINE__) ")\n"); }
#undef CCC
#define CCC

/* $Id: process.c,v 1.7 1999/09/23 00:05:41 gniibe Exp $
 *
 *  linux/arch/sh/kernel/process.c
 *
 *  Copyright (C) 1995  Linus Torvalds
 *
 *  SuperH version:  Copyright (C) 1999  Niibe Yutaka & Kaz Kojima
 */

/*
 * This file handles the architecture-dependent parts of process handling..
 */

#define __KERNEL_SYSCALLS__
#include <stdarg.h>

#include <linux/errno.h>
#include <linux/sched.h>
#include <linux/kernel.h>
#include <linux/mm.h>
#include <linux/smp.h>
#include <linux/smp_lock.h>
#include <linux/stddef.h>
#include <linux/ptrace.h>
#include <linux/malloc.h>
#include <linux/vmalloc.h>
#include <linux/user.h>
#include <linux/a.out.h>
#include <linux/interrupt.h>
#include <linux/unistd.h>
#include <linux/delay.h>
#include <linux/reboot.h>
#include <linux/init.h>

#include <asm/uaccess.h>
#include <asm/pgtable.h>
#include <asm/system.h>
#include <asm/io.h>
#include <asm/processor.h>
#include <asm/mmu_context.h>
#include <asm/elf.h>
#include <asm/syscall.h>

#include <linux/irq.h>

typedef unsigned long OBJECT_NAME;
typedef unsigned char UBYTE;
typedef long BYTE4;
typedef struct Queue_pbox{
	UBYTE  *data;		/* $(B%G!<%?$X$N%]%$%s%?(B */
	BYTE4  size;		/* $(B%G!<%?$N%5%$%:(B */
	BYTE4  timeout;		/* $(B%?%$%`%"%&%H$N;~9o(B */
	OBJECT_NAME sender;	/* $(B%G!<%?$NAw$j<g$NL>A0(B */
	BYTE4  when;		/* $(BAw$j<g$,%G!<%?$r%-%e!<%$%s%0$7$?;~$N%7%9%F%`;~9o(B */
} QUEUE_PBOX;


void restore_regs(struct pt_regs* regp);
void save_regs(struct pt_regs* regp);

extern void self_sleep(int);
extern void self_end(void);
extern int  proc_fork(void*,int);

int procname_seq = 0;

#if defined(__SH4__)
struct task_struct *last_task_used_math = NULL;
#endif

static int hlt_counter=0;

#define HARD_IDLE_TIMEOUT (HZ / 3)

void disable_hlt(void)
{
//q printk("--> disable_hlt\n");
	hlt_counter++;
}

void enable_hlt(void)
{
//q printk("--> enable_hlt\n");
	hlt_counter--;
}

/*
 * The idle loop on a uniprocessor i386..
 */ 
void cpu_idle(void *unused)
{
int	i, j;

	self_setpriority(20000);
	/* endless idle loop with no priority at all */
	init_idle();
	current->priority = 0;
	current->counter = -100;

//q printk("--> cpu_idle\n");
i = j = 0;
	while (1) {
		while (!current->need_resched) {
#if 1201 //{
			if(++i >= 1000){
				i = 0;
//0214 				printk("    cpu_idle %d\n",++j);
			}
			self_sleep(0);
#else
			if (hlt_counter)
				continue;
			__sti();
			asm volatile("sleep" : : : "memory");
#endif // 1201 }
		}
//q printk("    cpu_idle -> schedule\n");
		schedule();
		check_pgt_cache();
	}
}

void machine_restart(char * __unused)
{ /* Need to set MMU_TTB?? */
//q printk("--> machine_restart\n");
	for(;;){
		self_sleep(3600);
//q printk("    machine_restart\n");
	}
}

void machine_halt(void)
{
//q printk("--> machine_halt\n");
	for(;;){
		self_sleep(3600);
//q printk("    machine_halt\n");
	}
}

void machine_power_off(void)
{
//q printk("--> machine_power_off\n");
	for(;;){
		self_sleep(3600);
//q printk("    machine_power_off\n");
	}
}

void show_regs(struct pt_regs * regs)
{
//q printk("--> show_regs\n");
#if 1220 * 0
	printk("\n");
	printk("PC  : %08lx SP  : %08lx SR  : %08lx TEA : %08lx\n",
	       regs->pc, regs->sp, regs->sr, ctrl_inl(MMU_TEA));
	printk("R0  : %08lx R1  : %08lx R2  : %08lx R3  : %08lx\n",
	       regs->regs[0],regs->regs[1],
	       regs->regs[2],regs->regs[3]);
	printk("R4  : %08lx R5  : %08lx R6  : %08lx R7  : %08lx\n",
	       regs->regs[4],regs->regs[5],
	       regs->regs[6],regs->regs[7]);
	printk("R8  : %08lx R9  : %08lx R10 : %08lx R11 : %08lx\n",
	       regs->regs[8],regs->regs[9],
	       regs->regs[10],regs->regs[11]);
	printk("R12 : %08lx R13 : %08lx R14 : %08lx\n",
	       regs->regs[12],regs->regs[13],
	       regs->regs[14]);
	printk("MACH: %08lx MACL: %08lx GBR : %08lx PR  : %08lx\n",
	       regs->mach, regs->macl, regs->gbr, regs->pr);
#endif
}

struct task_struct * alloc_task_struct(void)
{
	/* Get two pages */
#if 1214 * 1
	unsigned long	a;
//q printk("--> " __FUNCTION__ "\n");
        a = (__get_free_pages(GFP_KERNEL,2) + ~CURRENT_MASK) & CURRENT_MASK;
//q printk("    " __FUNCTION__ "new addr=%lx\n",a);
	return (struct task_struct *)a;
#else
	return (struct task_struct *)((a + ~CURRENT_MASK) & CURRENT_MASK);
	return (struct task_struct *) __get_free_pages(GFP_KERNEL,1);
#endif
}

void free_task_struct(struct task_struct *p)
{
//q printk("--> free_task_struct\n");
	free_pages((unsigned long) p, 1);
}

/*
 * Create a kernel thread
 */

/*
 * This is the mechanism for creating a new kernel thread.
 *
 * NOTE! Only a kernel-only process(ie the swapper or direct descendants
 * who haven't done an "execve()") should use this: it will work within
 * a system call from a "real" process, but the process memory space will
 * not be free'd until both the parent and the child have exited.
 */
int kernel_thread(int (*fn)(void *), void * arg, unsigned long flags)
{
	int	r;
//0214 nanka_misete_hex3("--> kernel_thread(fn,arg,flags)",fn,arg,flags);
//q printk("--> kernel_thread\n");
	//new_fn = fn;
	//new_arg = arg;
	current->thread.fn = fn;
	current->thread.arg = arg;
//q printk("    kernel_thread fn=%p arg=%p\n",fn,arg);
	r = do_fork(flags|CLONE_VM, 0, 0);
//q printk("    kernel_thread do_fork()=%x\n",r);
//q printk("    kernel_thread <--\n");
	return r;
}
#if 0
{	/* Don't use this in BL=1(cli).  Or else, CPU resets! */
	register unsigned long __sc0 __asm__ ("r0") = __NR_clone;
	register unsigned long __sc4 __asm__ ("r4") = (long) flags | CLONE_VM;
	register unsigned long __sc5 __asm__ ("r5") = 0;
	register unsigned long __sc8 __asm__ ("r8") = (long) arg;
	register unsigned long __sc9 __asm__ ("r9") = (long) fn;
	__asm__ __volatile__(
		"trapa	#0\n\t" 	/* Linux/SH system call */
		"tst	#0xff,r0\n\t"	/* child or parent? */
		"bf	1f\n\t"		/* parent - jump */
		"jsr	@r9\n\t"	/* call fn */
		" mov	r8,r4\n\t"	/* push argument */
		"mov	r0,r4\n\t"	/* return value to arg of exit */
		"mov	%2,r0\n\t"	/* exit */
		"trapa	#0\n"
		"1:"
		:"=z" (__sc0)
		:"0" (__sc0), "i" (__NR_exit),
		 "r" (__sc4), "r" (__sc5), "r" (__sc8), "r" (__sc9)
		:"memory");
	return __sc0;
}
#endif

/*
 * Free current thread data structures etc..
 */
void exit_thread(void)
{
#if 1212 //{
//q printk("--> exit_thread\n");
	/* nothing to do ... */
#else
#if defined(__sh3__)
	/* nothing to do ... */
#elif defined(__SH4__)
#if 0 /* for the time being... */
	/* Forget lazy fpu state */
	if (last_task_used_math == current) {
		set_status_register (SR_FD, 0);
		write_system_register (fpscr, FPSCR_PR);
		last_task_used_math = NULL;
	}
#endif
#endif
#endif // 1212 }
}

void flush_thread(void)
{
#if 1212 //{
//q printk("--> flush_thread <--\n");
#if 0
	for(;;){
		self_sleep(3600);
//q printk("    flush_thread\n");
	}
#endif
#else
#if defined(__sh3__)
	/* do nothing */
	/* Possibly, set clear debug registers */
#elif defined(__SH4__)
#if 0 /* for the time being... */
	/* Forget lazy fpu state */
	if (last_task_used_math == current) {
		set_status_register (SR_FD, 0);
		write_system_register (fpscr, FPSCR_PR);
		last_task_used_math = NULL;
	}
#endif
#endif
#endif // 1212 }
}

void release_thread(struct task_struct *dead_task)
{
	unsigned long	pn, cn, spn, qr, qw;
	char		cfn[32];

	if(dead_task->thread.procname == current->thread.procname){
		/*
		 * dead_task is created by vfork.
		 * vforked child process dead without execve,
		 * this code is executed.
		 */

		proc_stop(spn);
		proc_delete(dead_task->thread.sys_procname);
		
		return;
	}


	spn = dead_task->thread.sys_procname;
	pn = dead_task->thread.procname;
	cn = dead_task->thread.codename;
	sprintf(cfn,"F1:ZXLNPN%c%c.BIN",(pn/26)%26+'A',pn%26+'A');
//	sprintf(cfn,"F1:ZXLNPN%c%c.BIN",(pn/26)%26+'0',pn%26+'0');
	qr = syscall_queuename_r(pn);
	qw = syscall_queuename_w(pn);

	proc_stop(pn);
	proc_stop(spn);

	proc_delete(spn);
	proc_delete(pn);
	code_delete(cn);
	FileDelete(cfn);
	queue_delete(qr);
	queue_delete(qw);
	
}

/* Fill in the fpu structure for a core dump.. */
int dump_fpu(struct pt_regs *regs, elf_fpregset_t *r)
{
#if defined(__SH4__)
#if 0 /* for the time being... */
	    /* We store the FPU info in the task->thread area.  */
	    if (! (regs->sr & SR_FD)) {
		    memcpy (r, &current->thread.fpu, sizeof (*r));
		    return 1;
	    }
#endif
#endif
	return 0; /* Task didn't use the fpu at all. */
}


extern void proc_start(int);
extern void proc_stop(int);
extern void queue_read(int,void*);
extern void new_proc_init(struct task_struct **, struct task_struct *);
extern void kernel_userwait(int, int);

static struct task_struct	*new_task;
static int			old_procname;


static void
thread_start(void)
{
//0214 nanka_misete("--> " __FUNCTION__);
//q printk("--> " __FUNCTION__ "\n");
//q printk("    " __FUNCTION__ " old_procname=%p\n",old_procname);
//q printk("    " __FUNCTION__ " current=%p\n",current);
//q printk("    " __FUNCTION__ " procname=%x\n",current->thread.sys_procname);
//q printk("    " __FUNCTION__ " --> proc_stop\n");
	while(old_procname == 0){
		self_sleep(0);
	}
//0214 nanka_misete_hex("    " __FUNCTION__ "2",old_procname);
	proc_start(old_procname);
	proc_stop(current->thread.sys_procname);
//0214 nanka_misete_hex3("    " __FUNCTION__ "3",current,current->thread.fn,current->thread.arg);
//q printk("    " __FUNCTION__ " --> fn %p(%p)\n",current->thread.fn,current->thread.arg);
	(*current->thread.fn)(current->thread.arg);
//0214 nanka_misete("    " __FUNCTION__ " <--");
//q printk("    " __FUNCTION__ " <-- fn\n");
}

#define PARANOIA 0
static void
thread_proc(void)
{
	unsigned long	cc;
#if PARANOIA
	if((((unsigned long)new_task) & ~CURRENT_MASK) != 0){
		panic(__FILE__ "," __FUNCTION__);
	}
#endif
	cc = (unsigned long)new_task;
	((volatile struct task_struct*)cc)->thread.sys_procname = self_getname();
	//cc += INIT_TASK_SIZE;
	cc += ~CURRENT_MASK + 1;
	__asm__ __volatile__ ("mov %0,r15" :: "r" (cc));
	thread_start();
	self_end();
}


int copy_thread(int nr, unsigned long clone_flags, unsigned long usp,
		struct task_struct *p, struct pt_regs *regs)
{
	int	n;
//q printk("--> copy_thread\n");
//q printk("    copy_thread new_task=%p\n",p);
//0124 nanka_misete_hex3("    copy_thread",p,p->thread.fn,p->thread.arg);
	p->thread.exec_notify = 0;
	new_task = p;
	old_procname = 0;
	n = proc_fork(thread_proc,PROC_FORK_STACK_SIZE);
	proc_stop(old_procname = self_getname());
//q printk("    copy_thread proc_fork sys_procname=%x current=%p\n",n,p);
//q printk("    copy_thread <--\n");
	return 0;
}

/*
 * fill in the user structure for a core dump..
 */
void dump_thread(struct pt_regs * regs, struct user * dump)
{
/* changed the size calculations - should hopefully work better. lbt */
	dump->magic = CMAGIC;
	dump->start_code = current->mm->start_code;
	dump->start_data  = current->mm->start_data;
	dump->start_stack = regs->sp & ~(PAGE_SIZE - 1);
	dump->u_tsize = (current->mm->end_code - dump->start_code) >> PAGE_SHIFT;
	dump->u_dsize = (current->mm->brk + (PAGE_SIZE-1) - dump->start_data) >> PAGE_SHIFT;
	dump->u_ssize = (current->mm->start_stack - dump->start_stack +
			 PAGE_SIZE - 1) >> PAGE_SHIFT;
	/* Debug registers will come here. */

	dump->regs = *regs;

#if 0 /* defined(__SH4__) */
	/* FPU */
	memcpy (&dump->regs[EF_SIZE/4], &current->thread.fpu,
		sizeof (current->thread.fpu));
#endif
}

#if 1219 * 0
/*
 *	switch_to(x,y) should switch tasks from x to y.
 *
 */
void __switch_to(struct task_struct *prev, struct task_struct *next)
{
	/*
	 * Restore the kernel stack onto kernel mode register
	 *   	k4 (r4_bank1)
	 */
	asm volatile("ldc	%0,r4_bank"
		     : /* no output */
		     :"r" ((unsigned long)next+8192));
}
#endif

static int copy_user_mem(unsigned long from, unsigned long to,
			  unsigned long start, unsigned long end)
{
	unsigned long scratch;
	register unsigned long lognend = end & 0xfffffffc;
	register int ret;

#if 0
	while ((start & 3) && start <= end){
		scratch = dbg_mem_get1(from, start);
		if ((ret = dbg_mem_put1(to, start, scratch)) == -1)
			return ret;
		start++;
	}
	while (start <= lognend){
		scratch = dbg_mem_get4(from, start);
		dbg_mem_put4(to, start, scratch);
		start+=4;
	}
#endif
	while (start <= end){
		scratch = dbg_mem_get1(from, start);
		if ((ret = dbg_mem_put1(to, start, scratch)) == -1)
			return ret;
		start++;
	}

	return 0;
}

extern int proc_memarea(unsigned long n, unsigned long a[]);
extern int create_proc_by_code(OBJECT_NAME, OBJECT_NAME);
extern OBJECT_NAME name_generate(void);
extern int queue_create(OBJECT_NAME, int, int);
extern unsigned long dbg_reg_get(OBJECT_NAME, int);

/*
 *  r0:
 *  r1:
 *  r2:	
 *  r3:			
 *  r4:	#101 (MEMORY_SBRK)
 *  r5:	0
 *  r6:  
 *  r7:	
 *  r8:	top of data section
 *  r9:	queue name
 *  r10: data
 *  r11: proc name
 *
 *  0:	02 c3       	trapa	#2
 *  2:	65 e4       	mov	#101,r4
 *
 *  4:	83 65       	mov	r8,r5
 *  6:	08 35       	sub	r0,r5
 *
 *  8:	02 c3       	trapa	#2
 *  a:	32 e4       	mov	#50,r4
 *
 *  c:	93 65       	mov	r9,r5
 *  e:	a3 66       	mov	r10,r6
 *
 * 10:	04 e7       	mov	#4,r7
 * 12:	ff e8       	mov	#-1,r8
 *
 * 14:	00 c3       	trapa	#0
 * 16:	13 e4       	mov	#19,r4
 *
 * 18:	b3 65       	mov	r11,r5
 * 1a:	00 c3       	trapa	#0
 */


#if 1
unsigned long initsbrk_code[] = {
	0xe465c302, 0x35086583, 0xe432c302, 0x66a36593,
	0xe8ffe704, 0xe413c300, 0xc30065b3
};
#elif 0
unsigned long initsbrk_code[] = {
	0x0009e432, 0x66a36593,
	0xe8ffe704, 0xe413c300, 0xc30065b3
};
#elif 0
unsigned long initsbrk_code[] = {
	0xe4130009, 0xc30065b3, 0x0009affe
};
#elif 1
unsigned long initsbrk_code[] = {
	0x00090009, 0x0009affe
};
#endif

void
return_sys_fork(void* arg)
{
	QUEUE_PBOX qr_arg;
	unsigned long queue_name;
	unsigned long parent_name = current->thread.procname;
	unsigned long child_name;
	struct pt_regs regs;
	unsigned long sp;
	unsigned long scratch;
	int i;
	unsigned long mem_area[4];
	mm_segment_t old_fs;
	int ret;
	char gomi[100];
	int data;

	printk ("return sys fork\n");

	/* 0: stop parent */
	proc_stop(parent_name);

	/* 1: exec */
	child_name = gen_username(current->pid);
	create_userproc_by_code(current->thread.codename, child_name);

#if 1
	/* 1.5: initial sbrk */
	old_fs = get_fs();
	set_fs(USER_DS);

	queue_name = name_generate();
	qr_arg.data = (unsigned long)&scratch;
	qr_arg.size = sizeof scratch;
	qr_arg.timeout = -1;
	queue_create(queue_name, 1, sizeof scratch);
	sp = dbg_reg_get(child_name, 15);
	sp &= 0xfffffffc;
	data = sp;
	sp -= sizeof initsbrk_code;

	printk ("queue name = %p\n", queue_name);
	printk ("data = %p\n", data);

	current->thread.procname = child_name;
#if 0
	__copy_to_user(sp, initsbrk_code, sizeof initsbrk_code);
#else
	{
	  int i;
	  printk ("INITSBRK(%p): ", sp);
	  for (i = 0; i < sizeof initsbrk_code; i++){
	    printk ("%02X ", ((unsigned char*)initsbrk_code)[i]);
	    ret = dbg_mem_put1(child_name, sp+i, ((char*)initsbrk_code)[i]);
	    if (ret < 0)
	      printk ("error");
	  }
	  printk ("\n");
	}
#endif

	proc_memarea(parent_name, mem_area); 
	regs.regs[4] = 101;
	regs.regs[5] = 0;
	regs.regs[8] = mem_area[1];
	regs.regs[9] = queue_name;
	regs.regs[10] = data;
	regs.regs[11] = child_name;
	regs.pc = sp;
	restore_regs(&regs);

	set_fs(old_fs);

	proc_start(child_name);
	printk ("proc_start = %d\n", ret);
#if 0
	queue_read(queue_name,&qr_arg);
	printk ("queue_read = %d\n", ret);
	queue_delete(queue_name);
	printk ("queue_delete = %d\n", ret);
#endif
self_sleep(10);

	proc_stop(child_name);
	printk ("proc_stop = %d\n", ret);

	save_regs(&regs);
#else
	proc_memarea(parent_name, mem_area); 
#endif

	/* 2: copy */
	{
		int ret;
		ret = copy_user_mem (parent_name, child_name, mem_area[2], mem_area[3]);
		if (ret < 0)
			panic ("stack copy failed\n");
		ret = copy_user_mem (parent_name, child_name, mem_area[0], mem_area[1]);
		if (ret < 0)
			panic ("data copy failed\n");
	}
		
	current->thread.procname = parent_name;
	save_regs(&regs);
	current->thread.procname = child_name;
	restore_regs(&regs);

	/* 2.5: queue generate */
	queue_generate(child_name);

	/* 3: start parent and child*/
	proc_start(child_name);
	proc_start(parent_name);
	mm_release();

	kernel_userwait(current->thread.procname, 1);
}

asmlinkage int sys_fork(unsigned long r4, unsigned long r5,
			unsigned long r6, unsigned long r7,
			struct pt_regs regs)
{
	struct pt_regs regs;

	memset(&regs,0,sizeof(regs));
	current->thread.fn = return_sys_fork;
	current->thread.arg = 0;

	printk ("calling do_fork\n");
	/*
	 * call with a CLONE_VFORK in order not to return to the parent
	 * too early.
	 * we call a mm_release() in a return_sys_fork().
	 */
	return do_fork(CLONE_VFORK | SIGCHLD, regs.sp, &regs);
//	return sys_vfork();
}

asmlinkage int sys_clone(unsigned long clone_flags, unsigned long newsp,
			 unsigned long r6, unsigned long r7,
			 struct pt_regs regs)
{
	if (!newsp)
		newsp = regs.sp;
	return do_fork(clone_flags, newsp, &regs);
}

void
return_sys_vfork(void* arg)
{
CCC  
	kernel_userwait(current->thread.procname, 1);
}

/*
 * This is trivial, and on the face of it looks like it
 * could equally well be done in user mode.
 *
 * Not so, for quite unobvious reasons - register pressure.
 * In user mode vfork() cannot have a stack frame, and if
 * done by calling the "clone()" system call directly, you
 * do not have enough call-clobbered registers to hold all
 * the information you need.
 */
#if 1224 * 0
int sys_vfork(void)
{
	struct pt_regs regs;

	memset(&regs,0,sizeof(regs));
	current->thread.fn = return_sys_vfork;
	current->thread.arg = 0;
	return do_fork(CLONE_VFORK | CLONE_VM | SIGCHLD, regs.sp, &regs);
}
#else
asmlinkage int sys_vfork(unsigned long from)
{
	struct pt_regs regs;
	int ret;
	int procname = current->thread.procname;
	char* stack;
	int len;
	mm_segment_t old_fs;

	current->thread.fn = return_sys_vfork;
	current->thread.arg = 0;

	proc_stop(procname);
	save_regs(&regs);

	printk ("regs.sp = %p, from = %p\n", regs.sp, from);
	len = from - regs.sp + 4;
	from = regs.sp;
	stack = kmalloc(len, GFP_KERNEL);
	old_fs = get_fs();
	set_fs(USER_DS);
	ret = __copy_from_user(stack, from, len);
	if (ret)
		panic("copy_form_user(stack) failed");
	set_fs(old_fs);
	proc_start(procname);

	ret = do_fork(CLONE_VFORK | CLONE_VM | SIGCHLD, regs.sp, &regs);
	printk ("do_fork = %d\n", ret);
	if (ret > 0){
        	int ret;
		proc_stop(procname);
		restore_regs(&regs);
		printk ("stack copy to %p len = %p\n", from, len);
		old_fs = get_fs();
		set_fs(USER_DS);
		ret = __copy_to_user(from, stack, len);
		if (ret)
			panic("copy_to_user(stack) failed");
		proc_start(procname);
		set_fs(old_fs);
	}
	kfree(stack);
	printk ("do_fork = %d\n", ret);
	return ret;
}	
#endif

/*
 * sys_execve() executes a new program.
 */
int sys_execve(char *ufilename, char **uargv, char **uenvp)
{
	int error;
	char *filename;
	struct pt_regs regs;
	QUEUE_PBOX qpb;
	static unsigned long zero = 0;
	int retry;

	memset(&regs,0,sizeof(regs));
	lock_kernel();
	filename = getname(ufilename);
	error = PTR_ERR(filename);
	if (IS_ERR(filename))
		goto out;
	error = do_execve(filename, uargv, uenvp, &regs);
	if (error == 0){
		current->flags &= ~PF_DTRACE;
		putname(filename);

		if (current->thread.exec_notify){
			qpb.size = 4;
			qpb.timeout = -1;
			qpb.data = &current->thread.procname;
			retry = 5;
			while (queue_write(current->thread.exec_notify, &qpb) == -1 && retry--){
				self_sleep(10);
				queue_create(current->thread.exec_notify, 1, 4);
			}
		}

		unlock_kernel();
		return 0;
	}
	putname(filename);
out:

	if (current->thread.exec_notify){
		qpb.size = 4;
		qpb.timeout = -1;
		qpb.data = &zero;
		retry = 5;
		while (queue_write(current->thread.exec_notify, &qpb) == -1 && retry--){
			self_sleep(10);
			queue_create(current->thread.exec_notify, 1, 4);
		}
	}

	unlock_kernel();
	return error;
}


#if 1
/*
 * kernel_execve() executes a new program in kernel.
 */
int kernel_execve(char *ufilename, char **uargv, char **uenvp)
{
	int error;
	char *filename;
	struct pt_regs regs;

	memset(&regs,0,sizeof(regs));
//q printk("--> sys_execve ufilename = %s\n",ufilename);
	lock_kernel();
	filename = getname(ufilename);
//q printk("    sys_execve filename = %s\n",filename);
	error = PTR_ERR(filename);
//q printk("    sys_execve 1\n");
	if (IS_ERR(filename))
		goto out;
//q printk("    sys_execve 2\n");
	error = do_execve(filename, uargv, uenvp, &regs);
//q printk("    sys_execve 3 error=%x\n",error);
	if (error == 0){
		current->flags &= ~PF_DTRACE;
		putname(filename);
		unlock_kernel();
		kernel_userwait(current->thread.procname, 0);
	}
//q printk("    sys_execve 4\n");
	putname(filename);
//q printk("    sys_execve 5\n");
out:
//q printk("    sys_execve 6\n");
	unlock_kernel();
	return error;
}
#endif


void save_regs(struct pt_regs* regp)
{
	int i;

	for (i = 0; i < 15; i++)
		regp->regs[i] = dbg_reg_get(current->thread.procname, i);
	regp->sp = dbg_reg_get(current->thread.procname, 15);
	regp->sr = dbg_reg_get(current->thread.procname, 16);
	regp->gbr = dbg_reg_get(current->thread.procname, 17);
	regp->mach = dbg_reg_get(current->thread.procname, 18);
	regp->macl = dbg_reg_get(current->thread.procname, 19);
	regp->pr = dbg_reg_get(current->thread.procname, 20);
	regp->pc = dbg_reg_get(current->thread.procname, 21);

#if 1
 	printk ("save_regs:");
 	printk ("  r0=%p   r1=%p\n", regp->regs[0], regp->regs[1]);
 	printk ("  r2=%p   r3=%p\n", regp->regs[2], regp->regs[3]);
 	printk ("  r4=%p   r5=%p\n", regp->regs[4], regp->regs[5]);
 	printk ("  r6=%p   r7=%p\n", regp->regs[6], regp->regs[7]);
 	printk ("  r8=%p   r9=%p\n", regp->regs[8], regp->regs[9]);
 	printk ("  ra=%p   rb=%p\n", regp->regs[10], regp->regs[11]);
 	printk ("  rc=%p   rd=%p\n", regp->regs[12], regp->regs[13]);
 	printk ("  re=%p   sp=%p\n", regp->regs[14], regp->sp);
 	printk ("  sr=%p   gbr=%p\n", regp->sr, regp->gbr);
 	printk ("  mach=%p macl=%p\n", regp->mach, regp->macl);
 	printk ("  pr=%p   pc=%p\n", regp->pr, regp->pc);
#endif
}

void restore_regs(struct pt_regs* regp)
{
	int i;
#if 0214 * 0
	printk ("restore_regs:");
	printk ("  r0=%p   r1=%p\n", regp->regs[0], regp->regs[1]);
	printk ("  r2=%p   r3=%p\n", regp->regs[2], regp->regs[3]);
	printk ("  r4=%p   r5=%p\n", regp->regs[4], regp->regs[5]);
	printk ("  r6=%p   r7=%p\n", regp->regs[6], regp->regs[7]);
	printk ("  r8=%p   r9=%p\n", regp->regs[8], regp->regs[9]);
	printk ("  ra=%p   rb=%p\n", regp->regs[10], regp->regs[11]);
	printk ("  rc=%p   rd=%p\n", regp->regs[12], regp->regs[13]);
	printk ("  re=%p   sp=%p\n", regp->regs[14], regp->sp);
	printk ("  sr=%p   gbr=%p\n", regp->sr, regp->gbr);
	printk ("  mach=%p macl=%p\n", regp->mach, regp->macl);
	printk ("  pr=%p   pc=%p\n", regp->pr, regp->pc);
#endif
	for (i = 0; i < 15; i++)
		dbg_reg_put(current->thread.procname, i, regp->regs[i]);
	dbg_reg_put(current->thread.procname, 15, regp->sp);
	dbg_reg_put(current->thread.procname, 16, regp->sr);
	dbg_reg_put(current->thread.procname, 17, regp->gbr);
	dbg_reg_put(current->thread.procname, 18, regp->mach);
	dbg_reg_put(current->thread.procname, 19, regp->macl);
	dbg_reg_put(current->thread.procname, 20, regp->pr);
	dbg_reg_put(current->thread.procname, 21, regp->pc);

	{
	  struct pt_regs regs;
	  
	  save_regs (&regs);
	  regp = &regs;

#if 0
	  printk ("restore_regs check:");
	  printk ("  r0=%p   r1=%p\n", regp->regs[0], regp->regs[1]);
	  printk ("  r2=%p   r3=%p\n", regp->regs[2], regp->regs[3]);
	  printk ("  r4=%p   r5=%p\n", regp->regs[4], regp->regs[5]);
	  printk ("  r6=%p   r7=%p\n", regp->regs[6], regp->regs[7]);
	  printk ("  r8=%p   r9=%p\n", regp->regs[8], regp->regs[9]);
	  printk ("  ra=%p   rb=%p\n", regp->regs[10], regp->regs[11]);
	  printk ("  rc=%p   rd=%p\n", regp->regs[12], regp->regs[13]);
	  printk ("  re=%p   sp=%p\n", regp->regs[14], regp->sp);
	  printk ("  sr=%p   gbr=%p\n", regp->sr, regp->gbr);
	  printk ("  mach=%p macl=%p\n", regp->mach, regp->macl);
	  printk ("  pr=%p   pc=%p\n", regp->pr, regp->pc);
#endif
	}
}

int
syscall_separate(int syscall_nr, int retval)
{
	struct pt_regs regs;

#if 0
/* ignore all signals */
	flush_signals(current);
#else
	printk ("syscall = %d <SIGNAL CHECK>\n", syscall_nr);
	if (signal_pending(current)){
		int invoke_handler = 0;
		printk ("<SIGNAL EXISTS>\n");

		proc_stop(current->thread.procname);
		save_regs(&regs);
		regs.syscall_nr = syscall_nr;
		regs.syscall_ret = retval;
		invoke_handler = do_signal(&regs, NULL);

#if 1
		if (signal_pending(current))
		  printk ("more signals are in pending\n");
#endif

		restore_regs(&regs);
		proc_start(current->thread.procname);


		/* calling signal handler ->
		 *    wait for next system call without writing to return queue
		 */
		current->thread.user_state = 1;
		set_current_state(TASK_INTERRUPTIBLE);
		schedule();
		return invoke_handler;
	}
#endif
	current->thread.user_state = 1;
	set_current_state(TASK_INTERRUPTIBLE);
	schedule();

	/* no signal -> queue write and then call queue read */
	return 0;
}

int
syscall_wait(void)
{
	/*
	 * current->thread.user_state
	 * 0 ä
	 * 1 queue_read ǥ֥å. linuxǤ ASK_INTERRUPTIBLE
	 * 2 none
	 * 3 TASK_RUNNINGˤʤꤿʤ
	 * 4 TASK_RUNNING
	 */
	//set_current_state(TASK_INTERRUPTIBLE);
	//schedule();
	//queue_read(callq, cbox);
	//printk("<==== %x\n",callq);
#if 1225
CCC
  printk ("%s curren->state = %d\n", __FUNCTION__, current->state);
	wake_up_process(current);

  init_task.need_resched = 1;
  
CCC
#else
	current->state = TASK_RUNNING;
#endif
	current->thread.user_state = 3;
	proc_stop(current->thread.sys_procname);
	current->thread.user_state = 4;
}

int get_thread_procname(void)
{
	return current->thread.procname;
}

void proc_dead(void)
{
	proc_dead_segv();
}

#if 1223

int sys_setup(void)
{
//0214 	printk(__FUNCTION__ " not support\n");
	return -ENOSYS;
}

int sys_break(void)
{
//0214         printk(__FUNCTION__ " not support\n");
	return -ENOSYS;
}

int sys_oldstat(void)
{
//0214         printk(__FUNCTION__ " not support\n");
	return -ENOSYS;
}

int sys_oldfstat(void)
{
//0214         printk(__FUNCTION__ " not support\n");
	return -ENOSYS;
}

int sys_stty(void)
{
//0214         printk(__FUNCTION__ " not support\n");
	return -ENOSYS;
}

int sys_gtty(void)
{
//0214         printk(__FUNCTION__ " not support\n");
	return -ENOSYS;
}

int sys_ftime(void)
{
//0214         printk(__FUNCTION__ " not support\n");
	return -ENOSYS;
}

int sys_prof(void)
{
//0214         printk(__FUNCTION__ " not support\n");
	return -ENOSYS;
}

int sys_lock(void)
{
//0214         printk(__FUNCTION__ " not support\n");
	return -ENOSYS;
}

int sys_mpx(void)
{
//0214         printk(__FUNCTION__ " not support\n");
	return -ENOSYS;
}

int sys_ulimit(void)
{
//0214         printk(__FUNCTION__ " not support\n");
	return -ENOSYS;
}

int sys_oldolduname(void)
{
//0214         printk(__FUNCTION__ " not support\n");
	return -ENOSYS;
}

int sys_oldlstat(void)
{
//0214         printk(__FUNCTION__ " not support\n");
	return -ENOSYS;
}

int sys_readdir(void)
{
//0214         printk(__FUNCTION__ " not support\n");
	return -ENOSYS;
}

int sys_profil(void)
{
//0214         printk(__FUNCTION__ " not support\n");
	return -ENOSYS;
}

int sys_ioperm(void)
{
//0214         printk(__FUNCTION__ " not support\n");
	return -ENOSYS;
}

int sys_olduname(void)
{
//0214         printk(__FUNCTION__ " not support\n");
	return -ENOSYS;
}

int sys_iopl(void)
{
//0214         printk(__FUNCTION__ " not support\n");
	return -ENOSYS;
}

int sys_idle(void)
{
//0214         printk(__FUNCTION__ " not support\n");
	return -ENOSYS;
}

int sys_vm86old(void)
{
//0214         printk(__FUNCTION__ " not support\n");
	return -ENOSYS;
}

int sys_modify_ldt(void)
{
//0214         printk(__FUNCTION__ " not support\n");
	return -ENOSYS;
}

int sys_afs_syscall(void)
{
//0214         printk(__FUNCTION__ " not support\n");
	return -ENOSYS;
}

int sys__llseek(void)
{
//0214         printk(__FUNCTION__ " not support\n");
	return -ENOSYS;
}

int sys__newselect(void)
{
//0214         printk(__FUNCTION__ " not support\n");
	return -ENOSYS;
}

int sys__sysctl(void)
{
//0214         printk(__FUNCTION__ " not support\n");
	return -ENOSYS;
}

int sys_vm86(void)
{
//0214         printk(__FUNCTION__ " not support\n");
	return -ENOSYS;
}

int sys_streams1(void)
{
//0214         printk(__FUNCTION__ " not support\n");
	return -ENOSYS;
}

int sys_streams2(void)
{
//0214         printk(__FUNCTION__ " not support\n");
	return -ENOSYS;
}
#endif
