#define CCC1(x) #x
#define CCC2(x) CCC1(x)
#define CCC miru(__FUNCTION__ " (" __FILE__ ":" CCC2(__LINE__) ")\n");
//#undef CCC
//#define CCC
/*
 *  linux/init/main.c
 *
 *  Copyright (C) 1991, 1992  Linus Torvalds
 *
 *  GK 2/5/95  -  Changed to support mounting root fs via NFS
 *  Added initrd & change_root: Werner Almesberger & Hans Lermen, Feb '96
 *  Moan early if gcc is old, avoiding bogus kernels - Paul Gortmaker, May '96
 *  Simplified starting of init:  Michael A. Griffith <grif@acm.org> 
 */


#define __KERNEL_SYSCALLS__

#include <linux/config.h>
#include <linux/proc_fs.h>
#include <linux/unistd.h>
#include <linux/string.h>
#include <linux/ctype.h>
#include <linux/delay.h>
#include <linux/utsname.h>
#include <linux/ioport.h>
#include <linux/init.h>
#include <linux/smp_lock.h>
#include <linux/blk.h>
#include <linux/hdreg.h>
#include <linux/iobuf.h>
#include <linux/bootmem.h>

#include <asm/io.h>
#include <asm/bugs.h>

#ifdef CONFIG_PCI
#include <linux/pci.h>
#endif

#ifdef CONFIG_DIO
#include <linux/dio.h>
#endif

#ifdef CONFIG_ZORRO
#include <linux/zorro.h>
#endif

#ifdef CONFIG_MTRR
#  include <asm/mtrr.h>
#endif

#ifdef CONFIG_APM
#include <linux/apm_bios.h>
#endif

#ifdef CONFIG_MAC
extern void nubus_init(void);
#endif

#ifdef CONFIG_ISAPNP
#include <linux/isapnp.h>
#endif

#ifdef CONFIG_IRDA
#include <net/irda/irda_device.h>
#endif

/*
 * Versions of gcc older than that listed below may actually compile
 * and link okay, but the end product can have subtle run time bugs.
 * To avoid associated bogus bug reports, we flatly refuse to compile
 * with a gcc that is known to be too old from the very beginning.
 */
#if __GNUC__ < 2 || (__GNUC__ == 2 && __GNUC_MINOR__ < 6)
#error sorry, your GCC is too old. It builds incorrect kernels.
#endif

#if 1215 // { debug
extern void miru(char*,...);
extern void nanka_misete(char*);
extern void nanka_misete_hex(char*,...);
extern void nanka_misete_hex3(char*,...);
extern void nanka_misete_hex7(char*,...);
#endif // 1215 }

#if 1215 // { crystal
extern void self_sleep(int);
extern void self_end(void);
extern void self_setpriority(int);
extern void self_grab_cpu(void);
extern void self_ungrab_cpu(void);
extern int  self_getname(void);
extern int  proc_fork(void*,int);
#endif

#if 1215 // { syscall
extern long sys_dup(unsigned int);
#endif // 1215 }

extern char _stext, _etext;
extern char *linux_banner;

extern int console_loglevel;

static int init(void *);

extern void init_IRQ(void);
extern void init_modules(void);
extern void sock_init(void);
extern void fork_init(unsigned long);
extern void mca_init(void);
extern void sbus_init(void);
extern void ppc_init(void);
extern void sysctl_init(void);
extern void filescache_init(void);
extern void signals_init(void);
extern int init_pcmcia_ds(void);

extern void free_initmem(void);
extern void filesystem_setup(void);

extern void ecard_init(void);

#if defined(CONFIG_SYSVIPC)
extern void ipc_init(void);
#endif
#if defined(CONFIG_QUOTA)
extern void dquot_init_hash(void);
#endif

//1209
extern unsigned int sys_gettime(void);
//1222mk #define jiffies sys_gettime()

/*
 * Boot command-line arguments
 */
#define MAX_INIT_ARGS 8
#define MAX_INIT_ENVS 8

extern void time_init(void);

int rows, cols;

#ifdef CONFIG_BLK_DEV_INITRD
kdev_t real_root_dev;
#endif

int root_mountflags = MS_RDONLY;
char *execute_command = NULL;

static char * argv_init[MAX_INIT_ARGS+2] = { "init", NULL, };
static char * envp_init[MAX_INIT_ENVS+2] = { "HOME=/", "TERM=linux", NULL, };

/*
 * Read an int from an option string; if available accept a subsequent
 * comma as well.
 *
 * Return values:
 * 0 : no int in string
 * 1 : int found, no subsequent comma
 * 2 : int found including a subsequent comma
 */
int get_option(char **str, int *pint)
{
    char *cur = *str;

CCC
    if (!cur || !(*cur)) return 0;
    *pint = simple_strtol(cur,str,0);
    if (cur==*str) return 0;
    if (**str==',') {
        (*str)++;
        return 2;
    }

    return 1;
}

char *get_options(char *str, int nints, int *ints)
{
	int res,i=1;

CCC
    while (i<nints) {
        res = get_option(&str, ints+i);
        if (res==0) break;
        i++;
        if (res==1) break;
    }
	ints[0] = i-1;
	return(str);
}

static int __init profile_setup(char *str)
{
    int par;
CCC
    if (get_option(&str,&par)) prof_shift = par;
	return 1;
}

__setup("profile=", profile_setup);


static struct dev_name_struct {
	const char *name;
	const int num;
} root_dev_names[] __initdata = {
	{ "nfs",     0x00ff },
	{ "hda",     0x0300 },
	{ "hdb",     0x0340 },
	{ "hdc",     0x1600 },
	{ "hdd",     0x1640 },
	{ "hde",     0x2100 },
	{ "hdf",     0x2140 },
	{ "hdg",     0x2200 },
	{ "hdh",     0x2240 },
	{ "hdi",     0x3800 },
	{ "hdj",     0x3840 },
	{ "hdk",     0x3900 },
	{ "hdl",     0x3940 },
	{ "hdm",     0x5800 },
	{ "hdn",     0x5840 },
	{ "hdo",     0x5900 },
	{ "hdp",     0x5940 },
	{ "hdq",     0x5A00 },
	{ "hdr",     0x5A40 },
	{ "hds",     0x5B00 },
	{ "hdt",     0x5B40 },
	{ "sda",     0x0800 },
	{ "sdb",     0x0810 },
	{ "sdc",     0x0820 },
	{ "sdd",     0x0830 },
	{ "sde",     0x0840 },
	{ "sdf",     0x0850 },
	{ "sdg",     0x0860 },
	{ "sdh",     0x0870 },
	{ "sdi",     0x0880 },
	{ "sdj",     0x0890 },
	{ "sdk",     0x08a0 },
	{ "sdl",     0x08b0 },
	{ "sdm",     0x08c0 },
	{ "sdn",     0x08d0 },
	{ "sdo",     0x08e0 },
	{ "sdp",     0x08f0 },
	{ "ada",     0x1c00 },
	{ "adb",     0x1c10 },
	{ "adc",     0x1c20 },
	{ "add",     0x1c30 },
	{ "ade",     0x1c40 },
	{ "fd",      0x0200 },
	{ "md",      0x0900 },	     
	{ "xda",     0x0d00 },
	{ "xdb",     0x0d40 },
	{ "ram",     0x0100 },
	{ "scd",     0x0b00 },
	{ "mcd",     0x1700 },
	{ "cdu535",  0x1800 },
	{ "sonycd",  0x1800 },
	{ "aztcd",   0x1d00 },
	{ "cm206cd", 0x2000 },
	{ "gscd",    0x1000 },
	{ "sbpcd",   0x1900 },
	{ "eda",     0x2400 },
	{ "edb",     0x2440 },
	{ "pda",	0x2d00 },
	{ "pdb",	0x2d10 },
	{ "pdc",	0x2d20 },
	{ "pdd",	0x2d30 },
	{ "pcd",	0x2e00 },
	{ "pf",		0x2f00 },
	{ "apblock", APBLOCK_MAJOR << 8},
	{ "ddv", DDV_MAJOR << 8},
	{ NULL, 0 }
};

kdev_t __init name_to_kdev_t(char *line)
{
	int base = 0;
CCC
	if (strncmp(line,"/dev/",5) == 0) {
		struct dev_name_struct *dev = root_dev_names;
		line += 5;
		do {
			int len = strlen(dev->name);
			if (strncmp(line,dev->name,len) == 0) {
				line += len;
				base = dev->num;
				break;
			}
			dev++;
		} while (dev->name);
	}
	return to_kdev_t(base + simple_strtoul(line,NULL,base?10:16));
}

static int __init root_dev_setup(char *line)
{
CCC
	ROOT_DEV = name_to_kdev_t(line);
	return 1;
}

__setup("root=", root_dev_setup);

static int __init checksetup(char *line)
{
	struct kernel_param *p;

CCC
	p = &__setup_start;
	do {
		int n = strlen(p->str);
		if (!strncmp(line,p->str,n)) {
			if (p->setup_func(line+n))
				return 1;
		}
		p++;
	} while (p < &__setup_end);
	return 0;
}

/* this should be approx 2 Bo*oMips to start (note initial shift), and will
   still work even if initially too large, it will just take slightly longer */
unsigned long loops_per_sec = (1<<12);

/* This is the number of bits of precision for the loops_per_second.  Each
   bit takes on average 1.5/HZ seconds.  This (like the original) is a little
   better than 1% */
#define LPS_PREC 8


void __init calibrate_delay(void)
{
#if 0
	unsigned long ticks, loopbit;
	int lps_precision = LPS_PREC;

CCC
	loops_per_sec = (1<<12);

	printk("Calibrating delay loop... ");
	while (loops_per_sec <<= 1) {
		/* wait for "start of" clock tick */
		ticks = jiffies;
		while (ticks == jiffies)
			/* nothing */;
		/* Go .. */
		ticks = jiffies;
		__delay(loops_per_sec);
		ticks = jiffies - ticks;
		if (ticks)
			break;
	}

/* Do a binary approximation to get loops_per_second set to equal one clock
   (up to lps_precision bits) */
	loops_per_sec >>= 1;
	loopbit = loops_per_sec;
	while ( lps_precision-- && (loopbit >>= 1) ) {
		loops_per_sec |= loopbit;
		ticks = jiffies;
		while (ticks == jiffies);
		ticks = jiffies;
		__delay(loops_per_sec);
		if (jiffies != ticks)	/* longer than 1 tick */
			loops_per_sec &= ~loopbit;
	}

/* finally, adjust loops per second in terms of seconds instead of clocks */	
	loops_per_sec *= HZ;
#else
	loops_per_sec = (1<<6);
#endif
/* Round the value and print it */	
	printk("%lu.%02lu BogoMIPS\n",
		(loops_per_sec+2500)/500000,
		((loops_per_sec+2500)/5000) % 100);
}

static int __init readonly(char *str)
{
CCC
	if (*str)
		return 0;
	root_mountflags |= MS_RDONLY;
	return 1;
}

static int __init readwrite(char *str)
{
CCC
	if (*str)
		return 0;
	root_mountflags &= ~MS_RDONLY;
	return 1;
}

static int __init debug_kernel(char *str)
{
CCC
	if (*str)
		return 0;
	console_loglevel = 10;
	return 1;
}

__setup("ro", readonly);
__setup("rw", readwrite);
__setup("debug", debug_kernel);

/*
 * This is a simple kernel command line parsing function: it parses
 * the command line, and fills in the arguments/environment to init
 * as appropriate. Any cmd-line option is taken to be an environment
 * variable if it contains the character '='.
 *
 * This routine also checks for options meant for the kernel.
 * These options are not given to init - they are for internal kernel use only.
 */
static void __init parse_options(char *line)
{
	char *next;
	int args, envs;
CCC
	if (!*line){
		return;
	}
	args = 0;
	envs = 1;	/* TERM is set to 'linux' by default */
	next = line;
	while ((line = next) != NULL) {
		if ((next = strchr(line,' ')) != NULL)
			*next++ = 0;
		if (!strncmp(line,"init=",5)) {
			line += 5;
			execute_command = line;
			/* In case LILO is going to boot us with default command line,
			 * it prepends "auto" before the whole cmdline which makes
			 * the shell think it should execute a script with such name.
			 * So we ignore all arguments entered _before_ init=... [MJ]
			 */
			args = 0;
			continue;
		}
		if (checksetup(line))
			continue;
		
		/*
		 * Then check if it's an environment variable or
		 * an option.
		 */
		if (strchr(line,'=')) {
			if (envs >= MAX_INIT_ENVS)
				break;
			envp_init[++envs] = line;
		} else {
			if (args >= MAX_INIT_ARGS)
				break;
			argv_init[++args] = line;
		}
	}
	argv_init[args+1] = NULL;
	envp_init[envs+1] = NULL;
}


extern void setup_arch(char **);
extern void cpu_idle(void);

#ifndef __SMP__

#define smp_init()	do { } while (0)

#else

/* Called by boot processor to activate the rest. */
static void __init smp_init(void)
{
	/* Get other processors into their bootup holding patterns. */
	smp_boot_cpus();
	smp_threads_ready=1;
	smp_commence();
}		

#endif

/*
 *	Activate the first processor.
 */
#include <linux/interrupt.h>	// timer
#include <linux/tty.h>		// keybd
#include <linux/tty_flip.h>	// keybd
#include <linux/console.h>	// keybd
#include <linux/kbd_kern.h>	// keybd

void start_kernel(void);

extern int	root_mountflags;
extern struct	tty_driver console_driver;

extern int zaumsc_init(void);
extern int zausrc_init(void);

extern void zauview_init(int,...);
extern void zauview_write(char* buf, unsigned int count);
extern void zaucons_init(int,...);
extern void zaulog_init(void);
extern void zaulog_flush(void);
extern int zaucomm_init(void);
extern int zaucomm_putstr(char *str);
extern int zaucomm_putchar(unsigned int ch);
extern int zaucomm_getchar(void);
extern int zaucomm_gets(char* b, char* e);
extern int zaukey_getchar(void);
extern int do_cmd(int ac, char** av);
extern void InitHeap(int);
extern void zaurus_init(void);
extern void zaucons_putstr(char*);
extern void usercode_init(void);
extern void zxfb_serv(void);


#define PARANOIA 0
#define TIMER_SERV_DEBUG
#define USE_TIMER_PROC	0

#define MONMON	0


static void kernel_proc(void)
{
 	unsigned long	cc;
#if PARANOIA
	if((((unsigned long)&init_task) & ~CURRENT_MASK) != 0){
		panic(__FILE__ "," __FUNCTION__);
	}
#endif
	cc = (unsigned long)&init_task;
	((volatile struct task_struct*)cc)->thread.sys_procname = self_getname();
	cc += INIT_TASK_SIZE;
 	__asm__ __volatile__ ("mov %0,r15" :: "r" (cc));
	start_kernel();
	self_setpriority(20000);
 	cpu_idle();
}


#define TIMER_SERV_DEBUG

static void timer_serv(void)
{
	int		h, i, j;
	struct pt_regs	regs;
	  /**/
	//proc_setpriority(self_getname(),10000);
	self_setpriority(2500);
//	self_setpriority(20000);
	h = -10;
	i = j = 0;
miru(__FUNCTION__ "\n");
miru(__FUNCTION__ " &i=%p\n",&i);
miru(__FUNCTION__ " &j=%p\n",&j);
miru(__FUNCTION__ " &regs=%p\n",&regs);
	for(;;){
		do_timer(&regs);
#ifdef TIMER_SERV_DEBUG
#if 0
		printk ("TIMER_SERV(%d) ---%d--- current=%p\n", 
			i,
			self_getname() & 0x7FFFFFFF,
			current);
#endif
		if(++h >= 0){
			h = 0;
#if 0
			printk ("BH(%d) bh_mask=%08X bh_active=%08X active_bhs=%08X\n",
				h, bh_mask, bh_active, get_active_bhs());
#endif
			do_bottom_half();
			//printk ("BH exit\n");
		}
#endif
		if(++i >= 64*60){
			i = 0;
			printk(__FUNCTION__ " %d\n",++j);
		}
		self_sleep(1);
	}
}


#if USE_TIMER_PROC

union task_union timer_task
	__attribute__((__section__(".data.init_task")));

static void timer_proc(void)
{
	unsigned long	cc;
#ifdef PARANOIA
	if((((unsigned long)&timer_task) & ~CURRENT_MASK) != 0){
		panic(__FILE__ "," __FUNCTION__);
	}
#endif
	cc = (unsigned long)&timer_task;
	((volatile struct task_struct*)cc)->thread.sys_procname = self_getname();
	cc += INIT_TASK_SIZE;
	__asm__ __volatile__ ("mov %0,r15" :: "r" (cc));
	timer_serv();
}

#endif /* USE_TIMER_PROC */


static void kbd_serv(void)
{
	int	c;
	struct tty_struct	*tty;
	  /**/
miru(__FUNCTION__ "\n");
	self_setpriority(100);
	for(;;){
		c = zaukey_getchar();
miru(__FUNCTION__ " c=%d\n",c);
		tty = console_driver.table[fg_console];
		if(tty){
			tty_insert_flip_char(tty, c, 0);
			con_schedule_flip(tty);
		}
	}
}


#if MONMON
static char *str2arg(char *b, char *e, char **avb, char **ave, int *acp)
{
	int	c, cond;
	char	*p, *q, **avp;
	  /**/
	for(p = q = b, avp = avb, cond = 0;;){
		if(p >= e){
			p = 0;
			break;
		}
		c = *(unsigned char*)p++;
		if(c <= 32 || c == 127 || (c & 128) != 0){
			if(cond == 1){
				*q++ = 0;
				++avp;
				cond = 0;
			}else if(cond == 3){
				--avp;
				cond = 2;
			}
			if(c == 0 || c == '\n'){
				break;
			}
		}else{
			if(cond == 0){
				if(avp < ave){
					*avp = q;
					*q++ = c;
					cond = 1;
				}else{
					avp = avb - ave + avb;
					cond = 3;
				}
			}else if(cond == 1){
				*q++ = c;
			}else if(cond == 2){
				cond = 3;
			}
		}
	}
	*acp = avp - avb;
	return p;
}


#define COMM_ECHO 0

#if COMM_ECHO
static void zaucomm_putnum(int n)
{
	int	t, f;
	char	*p, b[16];
	  /**/
	p = b + sizeof(b);
	*--p = 0;
	f = 0;
	if(n < 0){
		f = 1;
		n = -n;
	}
	do{
		t = n / 10;
		*--p = (n - t * 10) + '0';
	}while((n = t) != 0);
	if(f != 0){
		*--p = '-';
	}
	zaucomm_putstr(p);
}

static void prarg(int ac, char **av)
{
	int	i;
	  /**/
	if(ac < 0){
		ac = sizeof(av)/sizeof(av[0]);
	}
	for(i = 0; i < ac; ++i){
		zaucomm_putchar((i/10)+'0');
		zaucomm_putchar((i%10)+'0');
		zaucomm_putchar(' ');
		zaucomm_putstr(av[i]);
		zaucomm_putstr("\015\012");
	}
}
#endif

static void comm_serv(void)
{
	int	c, ac;
	char	*p, *e, *av[16], b[256];
	struct tty_struct	*tty;
	  /**/
miru(__FUNCTION__"\n");
	self_setpriority(100);
	zaucomm_init();
	zaucomm_putstr(linux_banner);
	e = b + sizeof(b) - 1;
	for(;;){
		zaucomm_putstr("> ");
		c = zaucomm_gets(b,e);
		if(b[0] == '!'){
			for(p = b + 1; (c = *(unsigned char*)p++) != 0;){
				tty = console_driver.table[fg_console];
				if(tty){
					tty_insert_flip_char(tty, c, 0);
					con_schedule_flip(tty);
				}
			}
			tty = console_driver.table[fg_console];
			if(tty){
				tty_insert_flip_char(tty, 10, 0);
				con_schedule_flip(tty);
			}
			continue;
		}
		if(c == 10){
			p = str2arg(b,e,av,av+sizeof(av)/sizeof(av[0]),&ac);
#if COMM_ECHO
			if(p == 0){
				zaucomm_putchar('0');
			}else if(p == b){
				zaucomm_putchar('b');
			}else if(p == e){
				zaucomm_putchar('e');
			}else{
				zaucomm_putnum(p-b);
				zaucomm_putchar(':');
			}
			zaucomm_putnum(ac);
			zaucomm_putstr("\015\012");
			prarg(ac,av);
#endif
			if(ac > 0){
				do_cmd(ac,av);
			}
		}
	}
}
#endif /*MONMON*/


#define static
#include <video/font.h>
#include <../drivers/video/font_6x11.c>
/**/
#undef FONTDATAMAX
#include <../drivers/video/font_8x8.c>
/**/
#undef FONTDATAMAX
#include <../drivers/video/font_8x16.c>
/**/
#undef FONTDATAMAX
#undef static
#if 0
font_6x11.c		fontdata_6x11		(11*256)
font_8x16.c		fontdata_8x16		4096
font_8x8.c		fontdata_8x8		2048
font_acorn_8x8.c	acorndata_8x8
font_pearl_8x8.c	fontdata_pearl8x8	2048
font_sun12x22.c		fontdata_sun12x22	11264
font_sun8x16.c		fontdata_sun8x16	4096
#endif
unsigned char *linux_6x11      = fontdata_6x11;
unsigned char *linux_8x8       = fontdata_8x8;
unsigned char *linux_8x16      = fontdata_8x16;


static unsigned long proccount;
static unsigned long proctab[16];


void
more_stop()
{
	int			i;
	unsigned long		pn;
	struct task_struct	*p, *n;
	  /**/
	for(n = init_task.next_task; (p = n) != &init_task ;){
		n = p->next_task;
		release_thread(p);
miru(__FUNCTION__" p=%x\n",p);
	}
	for(i = 0; i < proccount; ++i){
		pn = proctab[i];
		if(pn == self_getname()){
			continue;
		}
miru(__FUNCTION__" i=%d pn=%x\n",i,pn);
		proc_stop(pn);
		proc_delete(pn);
	}
}


void
more_main(void)
{
	unsigned long	pn;
{ extern int edata[], end[]; int *p; for(p = edata; p < end; ++p) *p = 0; }
	InitHeap(0);
	//self_sleep(16);
	//self_grab_cpu();
	//proc_setslice(self_getname(),0);
	//proc_setpriority(self_getname(),2000);
	//-self_setpriority(1000);
	self_setpriority(2000);

	zaurus_init();
{ extern int edata[], end[]; miru("edata=%x end=%x\n",edata,end); }
	miru("%s",linux_banner);
	proccount = 0;

	/*
	 *	start zxfb
	 */
#if 0306
	pn = proc_fork(zxfb_serv,PROC_FORK_STACK_SIZE);
	miru("     more_main <-- proc_fork (zxfb_serv)\n");
	proctab[proccount++] = pn;
#else
	ul = startsysproc("F1:ZXFB.BIN");
	miru("startsysproc proc=%x\n",ul);
	self_sleep(60);
#endif

#if MONMON
	/*
	 *	start zaucomm (for nikomon)
	 */
	pn = proc_fork(comm_serv,PROC_FORK_STACK_SIZE);
	miru("     more_main <-- proc_fork (comm_serv)\n");
	proctab[proccount++] = pn;
#endif /*MONMON*/

	/*
	 *	start timer driver
	 */
#if USE_TIMER_PROC
 	pn = proc_fork(timer_proc,PROC_FORK_STACK_SIZE);
	miru("     more_main <-- proc_fork (timer_proc)\n");
#else /* USE_TIMER_PROC */
 	pn = proc_fork(timer_serv,PROC_FORK_STACK_SIZE);
	miru("     more_main <-- proc_fork (timer_serv)\n");
#endif /* USE_TIMER_PROC */
	proctab[proccount++] = pn;

	/*
	 *	start keyboard driver
	 */
	pn = proc_fork(kbd_serv,PROC_FORK_STACK_SIZE);
	miru("     more_main <-- proc_fork (kbd_serv)\n");
	proctab[proccount++] = pn;

	/*
	 *	start kernel
	 */
	pn = proc_fork(kernel_proc,PROC_FORK_STACK_SIZE);
	miru("     more_main <-- proc_fork (kernel_proc)\n");
	proctab[proccount++] = pn;
	
#if 1
	/*
	 *	start zaulog flush daemon
	 */
	proctab[proccount++] = self_getname();
	for(;;){
		self_sleep(16);
		zaulog_flush();
	}
#else
	self_end();
#endif
}


asmlinkage void __init start_kernel(void)
{
	char * command_line;
	unsigned long mempages;
/*
 * Interrupts are still disabled. Do necessary setups, then
 * enable them
 */

nanka_misete("--> start_kernel");
self_sleep(60);
nanka_misete_hex("    start_kernel current=",current);
	zaumsc_init();
//	zausrc_init();
	printk("start_kernel");
	printk(__FUNCTION__ " current=%p sizeof(struct task_struct)=%x\n",current,sizeof(struct task_struct));

	CCC
//============================================================
#if 0
{
extern int EnableAccessVRAM(void), DisableAccessVRAM(void);
printk("GetVramTop()=%p\n",GetVramTop());
printk("&EnableAccessVRAM()=%p\n",&EnableAccessVRAM);
printk("&DisableAccessVRAM()=%p\n",&DisableAccessVRAM);
self_sleep(600);
}
#endif
	//{ long *p; p = (long*)(((char*)current) + sizeof(struct task_struct)); *p = 0xbacab0c0; }
#if 0
{extern int syscall_table[];
	printk("syscall_table=%x\n",syscall_table);
	printk("syscall_table[0]=%x\n",syscall_table[0]);
	printk("syscall_table[1]=%x\n",syscall_table[1]);
	printk("syscall_table[20]=%x\n",syscall_table[20]);
	panic("##");
}
#endif
	lock_kernel();
	//printk(linux_banner);
zaucons_putstr(linux_banner);
zaucons_putstr("\r\n");

	setup_arch(&command_line);
	paging_init();
CCC
	trap_init();
	init_IRQ();
	sched_init();
	time_init();
	parse_options(command_line);
	/*
	 * HACK ALERT! This is early. We're enabling the console before
	 * we've done PCI setups etc, and console_init() must be aware of
	 * this. But we do want output early, in case something goes wrong.
	 */
	console_init();
#ifdef CONFIG_MODULES
//1226- printk("--- init_modules\n");
	init_modules();
#endif
	printk("prof_shift=%lx\n",prof_shift);
	if (prof_shift) {
		unsigned int size;
		/* only text is profiled */
		prof_len = (unsigned long) &_etext - (unsigned long) &_stext;
		prof_len >>= prof_shift;
		
		size = prof_len * sizeof(unsigned int) + PAGE_SIZE-1;
		prof_buffer = (unsigned int *) alloc_bootmem(size);
		memset(prof_buffer, 0, size);
	}

	kmem_cache_init();
	sti();
	self_ungrab_cpu();
	calibrate_delay();
	self_grab_cpu();
#ifdef CONFIG_BLK_DEV_INITRD
	// FIXME, use the bootmem.h interface.
	if (initrd_start && !initrd_below_start_ok && initrd_start < memory_start) {
		printk(KERN_CRIT "initrd overwritten (0x%08lx < 0x%08lx) - "
		    "disabling it.\n",initrd_start,memory_start);
		initrd_start = 0;
	}
#endif
	mem_init();
	kmem_cache_sizes_init();
#ifdef CONFIG_PROC_FS
	proc_root_init();
#endif
	mempages = num_physpages;

	fork_init(mempages);
	filescache_init();
	dcache_init();
	vma_init();
	buffer_init(mempages);
	page_cache_init(mempages);
	kiobuf_init();
	signals_init();
	inode_init();
	file_table_init();
#if defined(CONFIG_SYSVIPC)
	ipc_init();
#endif
#if defined(CONFIG_QUOTA)
	dquot_init_hash();
#endif
	check_bugs();
	printk("POSIX conformance testing by UNIFIX\n");

	/* 
	 *	We count on the initial thread going ok 
	 *	Like idlers init is an unlocked kernel thread, which will
	 *	make syscalls (and thus be locked).
	 */
//1226- printk("smp_init\n");
	smp_init();
	//CCC
	//memcpy(current,&init_task,sizeof(struct task_struct));
	//CCC
	kernel_thread(init, NULL, CLONE_FS | CLONE_FILES | CLONE_SIGHAND);
	unlock_kernel();
	current->need_resched = 1;
	
 	//cpu_idle();
}

#ifdef CONFIG_BLK_DEV_INITRD
static int do_linuxrc(void * shell)
{
	static char *argv[] = { "linuxrc", NULL, };

//qCCC
	close(0);close(1);close(2);
	setsid();
	(void) open("/dev/console",O_RDWR,0);
	(void) dup(0);
	(void) dup(0);
	return execve(shell, argv, envp_init);
}

static int __init no_initrd(char *s)
{
//qCCC
	mount_initrd = 0;
	return 1;
}

__setup("noinitrd", no_initrd);

#endif

struct task_struct *child_reaper = &init_task;

static void __init do_initcalls(void)
{
	initcall_t *call;

//qCCC
	call = &__initcall_start;
	do {
		printk ("[[%p] = %p]", call, *call);
		(*call)();
		call++;
		printk("\n");
	} while (call < &__initcall_end);
	printk ("initcall end\n");
}

/*
 * Ok, the machine is now initialized. None of the devices
 * have been touched yet, but the CPU subsystem is up and
 * running, and memory and process management works.
 *
 * Now we can finally start doing some real work..
 */
static void __init do_basic_setup(void)
{
#ifdef CONFIG_BLK_DEV_INITRD
	int real_root_mountflags;
#endif

//qCCC
	/*
	 * Tell the world that we're going to be the grim
	 * reaper of innocent orphaned children.
	 *
	 * We don't want people to have to make incorrect
	 * assumptions about where in the task array this
	 * can be found.
	 */
	child_reaper = current;

#if defined(CONFIG_MTRR)	/* Do this after SMP initialization */
/*
 * We should probably create some architecture-dependent "fixup after
 * everything is up" style function where this would belong better
 * than in init/main.c..
 */
	mtrr_init();
#endif

#ifdef CONFIG_SYSCTL
	sysctl_init();
#endif

	/*
	 * Ok, at this point all CPU's should be initialized, so
	 * we can start looking into devices..
	 */
#ifdef CONFIG_PCI
	pci_init();
#endif
#ifdef CONFIG_SBUS
	sbus_init();
#endif
#if defined(CONFIG_PPC)
	ppc_init();
#endif
#ifdef CONFIG_MCA
	mca_init();
#endif
#ifdef CONFIG_ARCH_ACORN
	ecard_init();
#endif
#ifdef CONFIG_ZORRO
	zorro_init();
#endif
#ifdef CONFIG_DIO
	dio_init();
#endif
#ifdef CONFIG_MAC
	nubus_init();
#endif
#ifdef CONFIG_ISAPNP
	isapnp_init();
#endif

	/* Networking initialization needs a process context */ 
	sock_init();

#ifdef CONFIG_BLK_DEV_INITRD
	real_root_dev = ROOT_DEV;
	real_root_mountflags = root_mountflags;
	if (initrd_start && mount_initrd) root_mountflags &= ~MS_RDONLY;
	else mount_initrd =0;
#endif

	do_initcalls();

	CCC
	/* .. filesystems .. */
	filesystem_setup();

	CCC
#ifdef CONFIG_IRDA
	irda_device_init(); /* Must be done after protocol initialization */
#endif
	CCC
#ifdef CONFIG_PCMCIA
	init_pcmcia_ds();		/* Do this last */
#endif
	CCC
	/* Mount the root filesystem.. */
//1226- printk("------------------------------------\nroot_mountflags=%x\n",root_mountflags);
//qCCC
	mount_root();
//1226- printk("-- mount root done -----------------\n");
//qCCC

#ifdef CONFIG_BLK_DEV_INITRD
	root_mountflags = real_root_mountflags;
	if (mount_initrd && ROOT_DEV != real_root_dev
	    && MAJOR(ROOT_DEV) == RAMDISK_MAJOR && MINOR(ROOT_DEV) == 0) {
		int error;
		int i, pid;

		pid = kernel_thread(do_linuxrc, "/linuxrc", SIGCHLD);
		if (pid>0)
			while (pid != wait(&i));
		if (MAJOR(real_root_dev) != RAMDISK_MAJOR
		     || MINOR(real_root_dev) != 0) {
			error = change_root(real_root_dev,"/initrd");
			if (error)
				printk(KERN_ERR "Change root to /initrd: "
				    "error %d\n",error);
		}
	}
#endif
}

char readbuf[512];

extern int sys_read(int,char*,int);
extern int sys_write(int,char*,int);
extern int kernel_execve(char*,char**,char**);

void
nanka_type(char *name)
{
	int	i, j;
nanka_misete("#### nanka_type");
//1226- printk("nanka_type %s\n",name);
	if((i = sys_open(name, O_RDONLY, 0)) >= 0){
		memset(readbuf,0,sizeof(readbuf));
		j = sys_read(i,readbuf,sizeof(readbuf));
		printk("read_size=%x sys_read=%x\n",sizeof(readbuf),j);
		printk("%s\n",readbuf);
		j = sys_close(i);
		printk("sys_close=%x\n",j);
	}
}

void xpmread(char *filename, int x, int y);

static int init(void * unused)
{
	int	j;
nanka_misete("#### INIT init INIT ####");
#if 0
	{
		for(i = 0; i < 3; ++i){
			/*+*/prptr("#### INIT ",i);
			self_sleep(120);
		}
	}
#endif
	lock_kernel();
nanka_misete("#### INIT #### do_basic_setup");
	do_basic_setup();

	/*
	 * Ok, we have completed the initial bootup, and
	 * we're essentially up and running. Get rid of the
	 * initmem segments and start the user-mode stuff..
	 */
self_sleep(60);
nanka_misete("#### INIT #### free_initmem");
	free_initmem();
nanka_misete("#### INIT #### free_initmem 2");

#if 1000 * 0
{
	unsigned long ul;
	self_sleep(60);
	ul = startsysproc("F1:ZXFB.BIN");
	nanka_misete_hex("    startsysproc proc=",ul);
	self_sleep(1);
}
nanka_misete("#### INIT #### parapara");
	self_sleep(1);
for(;;){
xpmread("/share/icon/dpenguin/dpenguin.xpm.0",400,400);
xpmread("/share/icon/dpenguin/dpenguin.xpm.1",400,400);
xpmread("/share/icon/dpenguin/dpenguin.xpm.2",400,400);
xpmread("/share/icon/dpenguin/dpenguin.xpm.3",400,400);
xpmread("/share/icon/dpenguin/dpenguin.xpm.4",400,400);
xpmread("/share/icon/dpenguin/dpenguin.xpm.5",400,400);
xpmread("/share/icon/dpenguin/dpenguin.xpm.6",400,400);
xpmread("/share/icon/dpenguin/dpenguin.xpm.7",400,400);
xpmread("/share/icon/dpenguin/dpenguin.xpm.8",400,400);
xpmread("/share/icon/dpenguin/dpenguin.xpm.9",400,400);
xpmread("/share/icon/dpenguin/dpenguin.xpm.10",400,400);
xpmread("/share/icon/dpenguin/dpenguin.xpm.11",400,400);
xpmread("/share/icon/dpenguin/dpenguin.xpm.12",400,400);
xpmread("/share/icon/dpenguin/dpenguin.xpm.13",400,400);
xpmread("/share/icon/dpenguin/dpenguin.xpm.14",400,400);
xpmread("/share/icon/dpenguin/dpenguin.xpm.15",400,400);
xpmread("/share/icon/dpenguin/dpenguin.xpm.16",400,400);
xpmread("/share/icon/dpenguin/dpenguin.xpm.17",400,400);
xpmread("/share/icon/dpenguin/dpenguin.xpm.18",400,400);
xpmread("/share/icon/dpenguin/dpenguin.xpm.19",400,400);
xpmread("/share/icon/dpenguin/dpenguin.xpm.20",400,400);
xpmread("/share/icon/dpenguin/dpenguin.xpm.21",400,400);
xpmread("/share/icon/dpenguin/dpenguin.xpm.22",400,400);
xpmread("/share/icon/dpenguin/dpenguin.xpm.23",400,400);
}
#endif

	unlock_kernel();

#if 0
nanka_misete("#### INIT #### open /etc/rc");
	if ((i = sys_open("/etc/rc", O_RDWR, 0)) < 0){
nanka_misete("#### INIT #### open /etc/rc sippai");
	}else{
nanka_misete("#### INIT #### open /etc/rc seikou");
		memset(readbuf,0,sizeof(readbuf));
		j = sys_read(i,readbuf,sizeof(readbuf));
		printk("read_size=%x sys_read=%x\n",sizeof(readbuf),j);
		printk("%s\n",readbuf);
		j = sys_close(i);
		printk("sys_close=%x\n",j);
	}

nanka_misete("#### INIT #### open /etc/omake");
	if ((i = sys_open("/etc/omake", O_RDWR, 0)) < 0){
nanka_misete("#### INIT #### open /etc/omake (write) sippai");
	}else{
nanka_misete("#### INIT #### open /etc/omake (write) seikou");
		j = sys_write(i,readbuf,strlen(readbuf));
		printk("write=%x sys_read=%x\n",strlen(readbuf),j);
		printk("%s\n",readbuf);
		j = sys_close(i);
		printk("sys_close=%x\n",j);
	}

nanka_misete("#### INIT #### open /etc/omake");
	if ((i = sys_open("/etc/omake", O_RDWR, 0)) < 0){
nanka_misete("#### INIT #### open /etc/omake (read) sippai");
	}else{
nanka_misete("#### INIT #### open /etc/omake (read) seikou");
		memset(readbuf,0,sizeof(readbuf));
		j = sys_read(i,readbuf,sizeof(readbuf));
		printk("read_size=%x sys_read=%x\n",sizeof(readbuf),j);
		printk("%s\n",readbuf);
		j = sys_close(i);
		printk("sys_close=%x\n",j);
	}
#endif

nanka_misete("#### INIT #### open console");
	if((j = sys_open("/dev/console", O_RDWR, 0)) < 0)
		printk("Warning: unable to open an initial console.\n");
#if 1
nanka_misete("#### INIT #### console write");
	printk("console %d\n",j);
	j=5;//j = sys_write(0,"yaho\n",5);
	printk("console write() %d\n",j);
nanka_misete("#### INIT #### console write done");
#endif

#if 0
nanka_misete("#### INIT #### dup 0 (1)");
	(void) sys_dup(0);
nanka_misete("#### INIT #### dup 0 (2)");
	(void) sys_dup(0);
#endif
	j = sys_dup(0);
	printk("dup(0)=%d\n",j);
	j = sys_dup(0);
	printk("dup(0)=%d\n",j);

#if 1
self_sleep(30);
nanka_misete("#### INIT #### exec ");
	kernel_execve("/sbin/init",argv_init,envp_init);
self_sleep(30);
#endif

#if 1
self_sleep(30);
nanka_misete("#### INIT #### echo ");
	j = sys_write(0,"yaho\n",5);
	printk("console write() %d\n",j);
	j = sys_write(1,"yaho\n",5);
	printk("console write() %d\n",j);
	j = sys_write(2,"yaho\n",5);
	printk("console write() %d\n",j);
	for(;;){
                int n;
                n = sys_read(0, readbuf, 1);
                printk("console read n=%d, '%c'\n", n, readbuf[0]);

                n = sys_write(1, readbuf, 1);
                printk("console write n=%d, '%c'\n", n, readbuf[0]);
/*
                {
                        int r, x;
                        r = get_user(x, &dbg_count);
                        printk("dbg_count=%d, r=%d\n", x, r);
                }
*/
        }
panic("#############################################");
#endif


#if 1221 * 0
	nanka_type("/sbin/badmagic");
self_sleep(30);
	kernel_execve("/sbin/badmagic",argv_init,envp_init);
self_sleep(30);

	nanka_type("/sbin/okmagic");
self_sleep(30);
	kernel_execve("/sbin/okmagic",argv_init,envp_init);
self_sleep(30);

	nanka_type("/sbin/init");
self_sleep(30);
	kernel_execve("/sbin/init",argv_init,envp_init);
self_sleep(30);

	nanka_type("/sbin/drawstr");
self_sleep(30);
	kernel_execve("/sbin/drawstr",argv_init,envp_init);
self_sleep(30);
#endif

#if 1221 * 1
	for(;;){
		int	i;
		self_sleep(120);
		printk("chedule %d\n",i);
		schedule();
	}
#endif

	/*
	 * We try each of these until one succeeds.
	 *
	 * The Bourne shell can be used instead of init if we are 
	 * trying to recover a really broken machine.
	 */

nanka_misete("#### INIT #### execute_command");
nanka_misete("execute_command");
	if (execute_command)
		kernel_execve(execute_command,argv_init,envp_init);
CCC
	kernel_execve("/sbin/init",argv_init,envp_init);
	kernel_execve("/etc/init",argv_init,envp_init);
	kernel_execve("/bin/init",argv_init,envp_init);
	kernel_execve("/bin/sh",argv_init,envp_init);
	panic("No init found.  Try passing init= option to kernel.");
}

#if 0
static char buf[10000];
static int  r[256], g[255], b[255];
static int  putbuf[640];
void xpmread(char *filename, int x, int y)
{
        int  fd;
        char *s;
	int status, c, i, j, width, height, color, nazo;

//	printk("xpmread kitayo! \n");

	if((fd= sys_open(filename,0,0)) < 0){
		panic(filename);
		exit(1);
	}

//	printk("xpmread open! \n");
	status = 0;

	memset(buf,0,sizeof(buf));
        if(sys_read(fd,buf,sizeof(buf)) <= 0){
		panic(filename);
	}else{
//	printk("xpmread naiyou! \n");
		s = buf;
		for(;;){
			if(*s == '"'){
				if(status == 0){
					status = 1;
					sscanf(s+1,"%d %d %d %d",&width,&height,&color,&nazo);
					i = 0;
					/*  printk("%d , %d , %d ,%d\n", width,height,color,nazo);   */
				}else if(status == 1){
					c = *(s+1);
					if(s+5 != "#"){
						r[c] = g[c] = b[c] = -1;
					}
					sscanf(s+6,"%4x%4x%4x",&r[c],&g[c],&b[c]);
					/*  printk("%c: r=%x, g=%x, b=%x\n",c,r[c],g[c],b[c]);  */
					if(++i == color){
						status = 2;
						i = 0;
					}
				}else{
					putbuf[0] = x - width;
					putbuf[1] = y + (i - height) * 2;
					putbuf[2] = width;
					putbuf[3] = 1;
					for(s=s+1,j=4; *s != '"'; s++,j++){
#define V5(x) ((x)>>11&31)
#define V6(x) ((x)>>10&63)
						c = *s;
						putbuf[j] = V5(r[c])<<11|V6(g[c])<<5|V5(b[c]);
						/* putbuf[j] = c; */
						/* printk("%c",c); */
						/*  printk("%d, r=%x, g=%x, b=%x\n",*s,r[c],g[c],b[c]);  */
					}
					fb_put(putbuf);
					/* printk("\n"); */
					self_sleep(0);
					i = i + 1;
				}
			}
			L1:
			if(*s == 0){
				break;
			}
			if(*s != '\n'){
				s = s + 1;
				goto L1;
			}
			s = s + 1;
		}
        }
        sys_close(fd);
//	printk("xpmread osimai! \n");
}
#endif
